from __future__ import print_function, absolute_import
import ConfigParser
import StringIO
from dns import resolver, reversename
from passman import Passman, PassmanException

DEBUG = True

DEFAULT_CONFIG = {
    'provider': 'local-config',
    'domain': 'local',
    'under_group': None,
    'reset_credential': '0',
    'use_credential': 'gateway',
    'username': None,
    'password': None,
    'ca': None,
    'check_host_name': '1'
}


class Plugin:
    def __init__(self, configuration):
        config = ConfigParser.SafeConfigParser(DEFAULT_CONFIG)
        config.readfp(StringIO.StringIO(configuration))

        self._use_credential = config.get('auth', 'use_credential')
        self._passman_provider = config.get('auth', 'provider')
        self._passman_domain = config.get('auth', 'domain')
        self._passman_username = config.get('auth', 'username')
        self._passman_password = config.get('auth', 'password')

        self._passman_group = config.get('search', 'under_group')

        self._passman = Passman(config)

    def remove_password_from_debug(self, kwargs):
        mod_kwargs = {}
        for key, value in kwargs.items():
            if 'password' in key:
                mod_kwargs[key] = '****'
            else:
                mod_kwargs[key] = value
        return mod_kwargs

    def get_password_list(self, session_id, cookie, target_username, target_host, **kwargs):
        if DEBUG:
            print('credstore get_password_list', target_username, target_host, self.remove_password_from_debug(kwargs))

        try:
            credential = self._get_credential(
                cookie, target_username, target_host, 'password', **kwargs)
            return self._send_response(cookie, passwords=[credential, ])
        except PassmanException as exc:
            print("Error checking out password for {}@{}: '{}'".format(target_username, target_host, exc))
            return self._send_response(cookie)

    def get_private_key_list(self, session_id, cookie, target_username, target_host, **kwargs):
        if DEBUG:
            print('credstore get_private_key_list', target_username, target_host,
                  self.remove_password_from_debug(kwargs))

        try:
            credential = self._get_credential(cookie, target_username, target_host, 'ssh-key', **kwargs)

            # ssh key type is hard-coded here...
            return self._send_response(cookie, private_keys=[('ssh-rsa', credential), ])
        except PassmanException as exc:
            print("Error checking out ssh-key for {}@{}: '{}'".format(target_username, target_host, exc))
            return self._send_response(cookie)

    def _get_credential(self, cookie, target_username, target_host, credential_type, **kwargs):
        self._auth_session(cookie, **kwargs)

        account_id = None
        if 'session_cookie' in kwargs:
            account_id = kwargs['session_cookie'].get('account', None)

        if not account_id:
            try:
                account = self._passman.get_account(target_username, target_host, self._passman_group)
            except PassmanException:
                if self._is_ip_address(target_host):
                    if DEBUG:
                        print('target host is ip address, try with fqdn')
                    target_fqdn = self._get_fqdn(target_host)
                    if DEBUG:
                        print('target fqdn:', target_fqdn)
                    account = self._passman.get_account(target_username, target_fqdn, self._passman_group)
                    if DEBUG:
                        print('Found account', account)
                else:
                    raise
        credential, is_exclusive = self._passman.checkout_credential(account, credential_type)
        if DEBUG:
            print("Found {} for {}@{}".format(credential_type, target_username, target_host))

        cookie['account'] = account['id']
        cookie['is_exclusive'] = is_exclusive
        return credential

    @staticmethod
    def _is_ip_address(target_host):
        # simplification
        return all([ch.isdigit() or ch == '.' for ch in target_host])

    @staticmethod
    def _get_fqdn(ip_address):
        addr = reversename.from_address(ip_address)
        fqdn = str(resolver.query(addr, "PTR")[0])
        if fqdn.endswith('.'):
            fqdn = fqdn[:-1]
        return fqdn.strip()

    def authentication_completed(self, session_id, cookie):
        return self._send_response(cookie)

    def session_ended(self, session_id, cookie):
        try:
            self._checkin_credentials(cookie)
        except PassmanException as exc:
            print("Error checking in credential {}".format(exc))
        return self._send_response(cookie)

    def _auth_session(self, cookie, **kwargs):
        access_token, auth_token = self._get_access_token(**kwargs)
        cookie['access_token'] = access_token
        cookie['auth_token'] = auth_token
        self._passman.set_access_token(access_token)
        self._passman.set_auth_token(auth_token)

    def _get_access_token(self, **kwargs):
        if 'session_cookie' in kwargs and 'token' in kwargs['session_cookie']:
            print('Access token is used for authentication')
            return kwargs['session_cookie']['token']

        print("no access token given, let's authenticate")
        username, password = self._get_passman_credential(**kwargs)
        return self._passman.authenticate(username, password, self._passman_provider, self._passman_domain)

    def _get_passman_credential(self, **kwargs):
        if self._use_credential == 'explicit':
            if self._passman_username is not None and self._passman_password is not None:
                print("Using explictly configured proxy user to authenticate")
                return self._passman_username, self._passman_password
            else:
                raise PassmanException('Missing username and password')
        elif self._use_credential == 'gateway':
            if 'gateway_username' in kwargs and 'gateway_password' in kwargs and \
                    kwargs['gateway_username'] and kwargs['gateway_password']:
                print("Using gateway user to authenticate")
                return kwargs['gateway_username'], kwargs['gateway_password']
            else:
                raise PassmanException('Missing gateway credential')
        else:
            raise PassmanException('Invalid credential source: {}'.format(self._use_credential))

    def _checkin_credentials(self, cookie):
        if 'is_exclusive' not in cookie or not cookie['is_exclusive']:
            return

        if 'account' not in cookie:
            raise PassmanException('Missing account')

        if 'access_token' not in cookie:
            raise PassmanException('Missing access token')

        self._passman.set_access_token(cookie['access_token'])
        print("Checkin credentials")
        self._passman.checkin_credentials(cookie['account'])

    @staticmethod
    def _send_response(cookie, passwords=None, private_keys=None):
        result = {
            'cookie': cookie
        }

        if passwords is not None:
            result['passwords'] = passwords

        if private_keys is not None:
            result['private_keys'] = private_keys
        return result


if __name__ == '__main__':
    import sys
    plugin = Plugin(open(sys.argv[1], 'r').read())

    session_id = 'session_id'
    if len(sys.argv) == 6:
        gw_user = sys.argv[4]
        gw_pass = sys.argv[5]
    else:
        gw_user = None
        gw_pass = None

    # method = plugin.get_private_key_list
    method = plugin.get_password_list

    result = method(target_host=sys.argv[2], target_username=sys.argv[3], session_id=session_id,
                    cookie={}, gateway_username=gw_user, gateway_password=gw_pass, session_cookie={})
    print('RESULT', result)
    result = plugin.authentication_completed(session_id=session_id, cookie=result['cookie'])
    result = plugin.session_ended(session_id=session_id, cookie=result['cookie'])
